
from enum import Enum


class BendDirection(Enum):

    NORTH   = 'North'
    EAST    = 'East'
    SOUTH   = 'South'
    WEST    = 'West'
    UNKNOWN = 'Unknown'
    NONE    = 'None'
    NOT_SET = 'Not Set'

